#!/usr/bin/env python

from .mongotimecurve import MongoTimeCurve, tz
from plantmeter.resource import *

import pymongo
from datetime import datetime

import unittest

def localDate(y,m,d):
    return tz.localize(datetime(y,m,d))

class Resource_Test(unittest.TestCase):

    def setUp(self):
        self.databasename = 'generationkwh_test'
        self.collection = 'production'

        self.connection = pymongo.Connection()
        self.connection.drop_database(self.databasename)
        self.db = self.connection[self.databasename]
        self.curveProvider = MongoTimeCurve(self.db, self.collection)

    def tearDown(self):
        self.connection.drop_database('generationkwh_test')

    def test_getEmpty(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        curveProvider = self.curveProvider 
        m = Meter(
                'meterName',
                'meterDescription',
                True,
                uri=uri,
                curveProvider = self.curveProvider)

        p = Plant('plantName','plantDescription',True)
        p.meters.append(m)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)

        self.assertEqual(
            list(vp.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
                2*25*[0] 
                )
      
    def test_update_onePlantOneMeter(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        curveProvider = self.curveProvider 
        m = Meter(
                'meterName',
                'meterDescription',
                True,
                uri=uri,
                curveProvider = self.curveProvider)

        p = Plant('plantName','plantDescription',True)
        p.meters.append(m)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)
        m.updateWh(
            localDate(2015,9,4),
            localDate(2015,9,5))

        self.assertEqual(
            list(vp.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,
                0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,
            ])

    def test_update_onePlantTwoMeters(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        curveProvider = self.curveProvider 
        m1 = Meter(
                'meterName1',
                'meterDescription1',
                True,
                uri=uri,
                curveProvider = self.curveProvider)

        m2 = Meter(
                'meterName2',
                'meterDescription2',
                True,
                uri=uri,
                curveProvider = self.curveProvider)

        p = Plant('plantName','plantDescription',True)
        p.meters.append(m1)
        p.meters.append(m2)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)
        m1.updateWh(localDate(2015,9,4),localDate(2015,9,5))
        m2.updateWh(localDate(2015,9,4),localDate(2015,9,5))

        self.assertEqual(
            list(vp.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,6,12,10, 8,16,34,68,24,24,10,6,2,0,0,0,0,0,
                0,0,0,0,0,0,0,0,8,14,12,10,18,36,70,26,26,12,8,4,0,0,0,0,0,
            ])

    def test_update_twoPlantsOneMeter(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        curveProvider = self.curveProvider 
        m1 = Meter(
                'meterName1',
                'meterDescription1',
                True,
                uri=uri,
                curveProvider = self.curveProvider)
        p1 = Plant('plantName1','plantDescription1',True)
        p1.meters.append(m1)
        p2 = Plant('plantName2','plantDescription2',True)

        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p1)
        vp.plants.append(p2)
        m1.updateWh(localDate(2015,9,4),localDate(2015,9,5))

        self.assertEqual(
            list(vp.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,
                0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,
            ])

    def test_update_twoPlantsTwoMeters(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        curveProvider = self.curveProvider 
        m1 = Meter(
                'meterName1',
                'meterDescription1',
                True,
                uri=uri,
                curveProvider = self.curveProvider)
        p1 = Plant('plantName1','plantDescription1',True)
        p1.meters.append(m1)

        m2 = Meter(
                'meterName2',
                'meterDescription2',
                True,
                uri=uri,
                curveProvider = self.curveProvider)
        p2 = Plant('plantName2','plantDescription2',True)
        p2.meters.append(m2)

        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p1)
        vp.plants.append(p2)
        m1.updateWh(localDate(2015,9,4),localDate(2015,9,5))
        m2.updateWh(localDate(2015,9,4),localDate(2015,9,5))

        self.assertEqual(
            list(vp.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,6,12,10, 8,16,34,68,24,24,10,6,2,0,0,0,0,0,
                0,0,0,0,0,0,0,0,8,14,12,10,18,36,70,26,26,12,8,4,0,0,0,0,0,
            ])


    def test_lastDate_empty(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = uri,
            curveProvider = self.curveProvider)
        p = Plant('plantName','plantDescription',True)
        p.meters.append(m)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)

        self.assertEqual(vp.lastMeasurementDate(), None)

    def test_lastDate_onePlantOneMeter(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = uri,
            curveProvider = self.curveProvider)
        p = Plant('plantName','plantDescription',True)
        p.meters.append(m)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)
        m.updateWh(localDate(2015,9,4), localDate(2015,9,5))

        self.assertEqual(vp.lastMeasurementDate(), localDate(2015,9,5))

    def test_lastDate_onePlantTwoMeters(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        m1 = Meter(
            'meterName1',
            'meterDescription1',
            True,
            uri = uri,
            curveProvider = self.curveProvider)
        uri = 'csv:/' + local_file('data/manlleu_20150804.csv')
        m2 = Meter(
            'meterName2',
            'meterDescription2',
            True,
            uri = uri,
            curveProvider = self.curveProvider)

        p = Plant('plantName','plantDescription',True)
        p.meters.append(m1)
        p.meters.append(m2)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p)
        m1.updateWh(localDate(2015,9,4), localDate(2015,9,5))
        m2.updateWh(localDate(2015,8,4), localDate(2015,8,5))

        self.assertEqual(vp.lastMeasurementDate(), localDate(2015,8,5))


    def test_lastDate_twoPlantsTwoMeters(self):
        uri = 'csv:/' + local_file('data/manlleu_20150904.csv')
        m1 = Meter(
            'meterName1',
            'meterDescription1',
            True,
            uri = uri,
            curveProvider = self.curveProvider)
        uri = 'csv:/' + local_file('data/manlleu_20150804.csv')
        m2 = Meter(
            'meterName2',
            'meterDescription2',
            True,
            uri = uri,
            curveProvider = self.curveProvider)

        p1 = Plant('plantName1','plantDescription1',True)
        p2 = Plant('plantName2','plantDescription2',True)
        p1.meters.append(m1)
        p2.meters.append(m2)
        vp = VirtualPlant('vplantName','vplantDescription',True)
        vp.plants.append(p1)
        vp.plants.append(p2)
        m1.updateWh(localDate(2015,9,4), localDate(2015,9,5))
        m2.updateWh(localDate(2015,8,4), localDate(2015,8,5))

        self.assertEqual(vp.lastMeasurementDate(), localDate(2015,8,5))


class Meter_Test(unittest.TestCase):
    def setUp(self):
        self.databasename = 'generationkwh_test'
        self.collection = 'production'

        self.connection = pymongo.Connection()
        self.connection.drop_database(self.databasename)
        self.db = self.connection[self.databasename]
        self.curveProvider = MongoTimeCurve(self.db, self.collection)
        self.uri = 'csv:/' + local_file('data/manlleu_20150904.csv')

    def tearDown(self):
        self.connection.drop_database('generationkwh_test')

    def test_get_whenEmpty(self):
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = self.uri,
            curveProvider = self.curveProvider)

        self.assertEqual(
            list(m.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            2*25*[0]
            )

    def test_get_afterUpdate(self):
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = self.uri,
            curveProvider = self.curveProvider)
        m.updateWh(localDate(2015,9,4), localDate(2015,9,5))

        self.assertEqual(
            list(m.getWh(
                localDate(2015,9,4),
                localDate(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,
                0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,
            ])

    def test_lastDate_empty(self):
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = self.uri,
            curveProvider = self.curveProvider)
        self.assertEqual(m.lastMeasurementDate(), None)

    def test_lastDate_filled(self):
        m = Meter(
            'meterName',
            'meterDescription',
            True,
            uri = self.uri,
            curveProvider = self.curveProvider)
        m.updateWh(localDate(2015,9,4), localDate(2015,9,5))

        self.assertEqual(m.lastMeasurementDate(), localDate(2015,9,5))


unittest.TestCase.__str__ = unittest.TestCase.id

# vim: et ts=4 sw=4



import datetime
import ftplib
from plantmeter.providers import BaseProvider, BaseProviderConnectionError, \
        BaseProviderDownloadError, register, urlparse
from plantmeter.utils import daterange

import unittest
from mock import patch

class MonsolProvider(BaseProvider):
    """Monsol Provider
    """

    def __init__(self, uri=None):
        if uri is None:
            uri = "monsol://path"
        super(MonsolProvider, self).__init__(uri)
        self.uri = uri
        self.res = urlparse(self.uri) 

    def extract(self, content, start, end=None):
        def parse(line):
            items = line.split(';')
            return {
                    'datetime': datetime.datetime.strptime(items[0], '%Y-%m-%d %H:%M'),
                    'daylight': items[1],
                    'ae': int(items[3])
                    }

        if not end:
            end = start
        return [[measure
            for date in daterange(start, end + datetime.timedelta(days=1))
                for measure in (parse(line) for line in content)
                    if measure['datetime'].date() == date
            ]]

    def download(self, filename):
        client = None
        try:
            client = ftplib.FTP(self.res['hostname'])
            client.login(self.res['username'], self.res['password'])
        except Exception as e:
            raise BaseProviderConnectionError(e)

        import StringIO
        from contextlib import closing
        with closing(StringIO.StringIO()) as sio:
            try:
                client.cwd(self.res['path'])
                client.retrbinary('RETR ' + filename, callback=sio.write)
            except Exception as e:
                raise BaseProviderDownloadError()
            finally:
                client.quit()

            sio.seek(0)
            return sio.readlines()

    def get(self, start, end=None):
        return [self.extract(date, self.download(date)) for date in daterange(start, end)]

    def __enter__(self):
        return self

register("monsol", MonsolProvider)


class MonsolProvide_Test(unittest.TestCase):
    def setUp(self):
        ""

    def tearDown(self):
        ""

    def load_file(self, path):
        with open(path, 'r') as f:
            output = f.readlines()
        return output

    def test_failHostname(self):
        m = MonsolProvider('monsol://user:password@hostname/path')
        self.assertRaises(BaseProviderConnectionError, m.download, 'tes.csv')

    def test_failConnection(self):
        from socket import error
        m = MonsolProvider('monsol://user:password@www.somenergia.coop/path')
        self.assertRaises(BaseProviderConnectionError, m.download, 'tes.csv')

    @patch('ftplib.FTP', autospec=True)
    def test_download(self, ftp_constructor):
        mock_ftp = ftp_constructor.return_value
        m = MonsolProvider('monsol://user:password@hostname/path')
        m.download('test.csv')
    
        # TODO: Implement callback mockup
        mock_ftp.login.assert_called_once_with('user', 'password')
        mock_ftp.cwd.assert_called_once_with('path')
        self.assertEqual(mock_ftp.retrbinary.call_args[0], ('RETR test.csv',))

    def test_extract(self):
        import os
        m = MonsolProvider('monsol://user:password@hostname/path')
        path = os.path.dirname(os.path.realpath(__file__)) + '/../data/monsol_20150904.csv'
        content = self.load_file(path)
        measurements = m.extract(content, datetime.date(2015,9,4))
        self.assertDictEqual(
                measurements[0][0],
                {
                    'daylight': 'S', 
                    'ae': 0, 
                    'datetime': datetime.datetime(2015, 9, 4, 0, 0)
                    })
        self.assertDictEqual(
                measurements[0][-1],
                {
                    'daylight': 'S',
                    'ae': 0,
                    'datetime': datetime.datetime(2015, 9, 4, 23, 0)
                    })

    @unittest.skip("tobe implemented!")
    def test_get(self):
        ""

# vim: et ts=4 sw=4



import os
import numpy as np

from plantmeter.providers import get_provider

"""
TODOs
- Ordre
- Duplicats
- Gaps
- Padding
"""

def local_file(filename):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)

class Resource(object):
    name = None
    description = None
    enabled = None

    def __init__(self, name, description, enabled):
        self.name = name
        self.description = description
        self.enabled = enabled


class VirtualPlant(Resource):
    def __init__(self, *args, **kwargs):
        self.plants = kwargs.pop('plants') if 'plants' in kwargs else []
        super(VirtualPlant, self).__init__(*args, **kwargs)

    def getWh(self, start, end):
        return np.sum([
            plant.getWh(start, end)
            for plant in self.plants
            if plant.enabled
            ], axis=0)

    def lastMeasurementDate(self):
        return min([
            plant.lastMeasurementDate()
            for plant in self.plants
            if plant.enabled
            ])

class Plant(Resource):
    def __init__(self, *args, **kwargs):
        self.meters = kwargs.pop('meters') if 'meters' in kwargs else []
        super(Plant, self).__init__(*args, **kwargs)

    def getWh(self, start, end):
        return np.sum([
            meter.getWh(start, end)
            for meter in self.meters
            if meter.enabled
            ], axis=0)

    def lastMeasurementDate(self):
        return min([
            meter.lastMeasurementDate()
            for meter in self.meters
            if meter.enabled
            ])

class Meter(Resource):
    def __init__(self, *args, **kwargs):
        self.uri = kwargs.pop('uri') if 'uri' in kwargs else None 
        self.curveProvider = kwargs.pop('curveProvider') if 'curveProvider' in kwargs else None
        super(Meter, self).__init__(*args, **kwargs)

    def getWh(self, start, end):
        return self.curveProvider.get(start, end, self.name, 'ae')

    def updateWh(self, start, end=None):
        provider = get_provider(self.uri)
        with provider(self.uri) as remote:
            for day in remote.get(start, end):
                for measurement in day:
                    self.curveProvider.fillPoint(
                            name = self.name,
                            datetime = measurement['datetime'],
                            ae = measurement['ae']
                            )
    def lastMeasurementDate(self):
        return self.curveProvider.lastFullDate(self.name)


# vim: et ts=4 sw=4



# -*- coding: utf-8 -*-

import os
from osv import osv, fields
from mongodb_backend import osv_mongodb
import netsvc
import pooler
import time
from tools.translate import _
from tools import config

from datetime import datetime
from plantmeter.resource import VirtualPlant, Plant, Meter 


class GenerationVirtualPlant(osv.osv):
    '''Implements generation virtual plant'''

    _name = 'generation.virtual.plant'

    def getWh(self, cursor, uid, pid, start, end, context=None):
        '''Get generation from virtual plant'''
   
        if not context:
            context = {}
        if isinstance(pid, list) or isinstance(pid, tuple):
            pid = pid[0]
        vplant = self.browse(cursor, uid, pid, context)

        def obj_to_dict(obj, attrs):
            return {attr: getattr(obj, attr) for attr in attrs}

        args = ['name', 'description', 'enabled']
        vplant = self.browse(cursor, uid, pid, context)

        _vplant = VirtualPlant(**obj_to_dict(vplant, args).update(
            {'plants': [Plant(**obj_to_dict(plant, args).update(
                {'meters': [Meter(**obj_to_dict(meter, args + ['uri']))
                for meter in plant.meters if meter.enabled]}))
            for plant in vplant.plants if plant.enabled]}))
        
        return _vplant.getWh(start, end)

    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'plants': fields.one2many('generation.plant', 'vplant_id', 'Plants')
    }

    _defaults = {
        'enabled': lambda *a: False
    }
GenerationVirtualPlant()


class GenerationPlant(osv.osv):

    _name = 'generation.plant'
    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'vplant_id': fields.many2one('generation.virtual.plant', 'Virtual Plant',
                                  required=True),
        'meters': fields.one2many('generation.meter', 'plant_id', 'Meters')
    }
    _defaults = {
        'enabled': lambda *a: False,
    }
GenerationPlant()


class GenerationMeter(osv.osv):

    _name = 'generation.meter'
    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'plant_id': fields.many2one('generation.plant'),
        'uri': fields.char('Host', size=150, required=True),
        }
    _defaults = {
        'enabled': lambda *a: False,
    }
GenerationMeter()


class GenerationMeasurement(osv_mongodb.osv_mongodb):

    _name = 'generation.measurement'
    _order = 'timestamp desc'

    def search(self, cursor, uid, args, offset=0, limit=0, order=None,
               context=None, count=False):
        '''Exact match for name.
        In mongodb, even when an index exists, not exact
        searches for fields scan all documents in collection
        making it extremely slow. Making name exact search
        reduces dramatically the amount of documents scanned'''

        new_args = []
        for arg in args:
            if not isinstance(arg, (list, tuple)):
                new_args.append(arg)
                continue
            field, operator, value = arg
            if field == 'name' and operator != '=':
                operator = '='
            new_args.append((field, operator, value))
        return super(GenerationMeasurement,
                     self).search(cursor, uid, new_args,
                                  offset=offset, limit=limit,
                                  order=order, context=context,
                                  count=count)

    _columns = {
        'name': fields.integer('Plant identifier'), # NOTE: workaround due mongodb backend
        'create_at': fields.datetime('Create datetime'),
        'datetime': fields.datetime('Exported datetime'),
        'daylight': fields.char('Exported datetime daylight',size=1),
        'ae': fields.float('Exported energy (kWh)')
    }
GenerationMeasurement()
